"""
Tables of named colors
"""

import itertools, re
from . import juce, classic


def canonical_name(name):
    return name.replace(' ', '').lower()


def to_triplet(color):
    rg, b = color // 256, color % 256
    r, g = rg // 256, rg % 256
    return r, g, b


def set_user_colors(colors):
    from . import names

    _TO_NAME_USER.clear()
    _TO_COLOR_USER.clear()

    for name, color in colors.items():
        color = names.to_color(color)
        _TO_NAME_USER[color] = name
        _TO_COLOR_USER[canonical_name(name)] = color


def get_color(name):
    name = canonical_name(name)
    return _TO_COLOR_USER.get(name) or _TO_COLOR.get(name)


def get_name(color):
    return _TO_NAME_USER.get(color) or _TO_NAME.get(color)


def all_named_colors():
    """Return an iteration over all name, color pairs in tables"""
    yield from _TO_COLOR_USER.items()
    for name, color in _TO_COLOR.items():
        if name not in _TO_COLOR_USER:
            yield name, color


def contains(x):
    """Return true if this string or integer tuple appears in tables"""
    if isinstance(x, str):
        x = canonical_name(x)
        return x in _TO_COLOR_USER or x in _TO_COLOR
    else:
        x = tuple(x)
        return x in _TO_NAME_USER or x in _TO_NAME


def _classic_pairs():
    find = re.compile(r'[A-Z][a-z0-9]+').finditer

    for name in dir(classic):
        if name == 'Green_HTML':
            yield 'green html', classic.Green_HTML

        elif len(name) > 1 and name[0].isupper() and name[1].islower():
            key = ' '.join(i.group(0).lower() for i in find(name))
            yield key, getattr(classic, name)


_JUCE_COLORS = {k: to_triplet(v) for k, v in juce.COLORS.items()}
_CLASSIC_COLORS = dict(_classic_pairs())

"""
A dictionary of every color by name.
"""

COLOR_DICT = dict(_CLASSIC_COLORS, **_JUCE_COLORS)
CANONICAL_DICT = {canonical_name(k): v for k, v in COLOR_DICT.items()}

_SECONDARY_NAMES = juce.SECONDARY_NAMES.union({'off', 'on'})
_TO_NAME = {v: k for k, v in COLOR_DICT.items() if k not in _SECONDARY_NAMES}
_TO_COLOR = {canonical_name(k): v for k, v in COLOR_DICT.items()}
_TO_NAME_USER = {}
_TO_COLOR_USER = {}



import time


class Clock:
    def __init__(self, is_flat_out=False):
        self.is_flat_out = is_flat_out
        self._time = time.time()

    def time(self):
        return self._time if self.is_flat_out else time.time()

    def sleep(self, delta_time):
        delta_time = max(delta_time, 0)
        # TODO: why is delta_time occasionally tiny and negative?
        if not self.is_flat_out:
            time.sleep(delta_time)
        else:
            self._time += delta_time



from . import tables


class Colors:
    """
    Colors is a "magic" color name object.

     To get a color from a name, use ``COLORS.<colorname>`` - for example

    ::

        COLORS.red
        COLORS.ochre

    or if the name is a variable or has a space in it,

    ::

        COLORS['violet red 4']

    To get a name from a color, use

    ::

        COLOR((0, 255, 255))
    """

    def __getitem__(self, name):
        value = tables.get_color(name)
        if value:
            return value
        raise KeyError(name)

    def __getattr__(self, name):
        value = tables.get_color(name)
        if value:
            return value
        raise AttributeError("COLORS has no attribute '%s'" % name)

    def __setitem__(self, name, value):
        raise KeyError('Cannot change COLORS')

    def __setattr__(self, name, value):
        raise AttributeError('Cannot change COLORS')

    def __iter__(self):
        return tables.all_named_colors()

    def __contains__(self, x):
        return tables.contains(x)


COLORS = Colors()



from . description import Description
from .. util import data_file


class SavedDescription:
    """
    Hold a Description and save and load it from a data_file.
    """
    def __init__(self, project_file='', desc=None, **kwds):
        self.desc = Description()

        self.project_file = project_file
        if project_file:
            self.load(project_file)
        else:
            self._project_file = project_file
        self.desc.update(desc, **kwds)

    def load(self, project_file=''):
        """Load/reload the description from a YML file. Prompt if no file given."""
        self._request_project_file(project_file)
        self.clear()
        self.desc.update(data_file.load(self._project_file))

    def save(self, project_file=''):
        """Save the description as a YML file. Prompt if no file given."""
        self._request_project_file(project_file)
        data_file.dump(self.desc.as_dict(), self.project_file)

    def clear(self):
        """Clear description to default values"""
        self.desc.clear()

    def __str__(self):
        return str(self.desc)

    def __repr__(self):
        rep = super().__repr__()
        if self.project_file:
            return rep + ' loaded from file ' + self.project_file
        else:
            return rep

    @property
    def project_file(self):
        return self._project_file

    @project_file.setter
    def project_file(self, project_file):
        if project_file and not project_file.endswith('.yml'):
            project_file += '.yml'
        self._project_file = project_file

    _ATTRIBUTES = 'project_file',

    def _request_project_file(self, project_file, action='store'):
        project_file = project_file or self.project_file
        if not project_file:
            project_file = input('Enter project_file to %s: ' % action).strip()
            if not project_file:
                raise ValueError('%s aborted', action.capitalize())
        self.project_file = project_file



import contextlib, os, signal
from . import log, pid_context

SIGNAL_NUMBERS = {k: v for k, v in signal.__dict__.items()
                  if k.startswith('SIG') and '_' not in k}

SIGNAL_NAMES = {v: k for k, v in SIGNAL_NUMBERS.items()}

STOP_SIGNALS = 'SIGINT', 'SIGTERM'
RESTART_SIGNALS = 'SIGHUP',
HANDLED_SIGNALS = STOP_SIGNALS + RESTART_SIGNALS


@contextlib.contextmanager
def context(**handlers):
    signals = {}

    def handler(signum, frame):
        signame = SIGNAL_NAMES[signum]
        signals[signame] = True
        handler = handlers[signame]
        handler()

    def set_all(handler):
        for signame in handlers:
            if signame in SIGNAL_NUMBERS:  # Windows doesn't have all signals
                signal.signal(SIGNAL_NUMBERS[signame], handler)

    set_all(handler)
    try:
        yield signals

    finally:
        set_all(signal.SIG_DFL)


def send_signal(sig, pid_filename=None):
    sig = SIGNAL_NUMBERS.get(sig, sig)

    if sig not in SIGNAL_NAMES:
        log.error('No signal %s', sig)
        return -1

    try:
        pid = pid_context.get_pid(pid_filename)
    except:
        log.error('No bp process running')
        log.debug('Could not find file %s', pid_filename)
        return -1

    try:
        os.kill(pid, sig)
    except:
        log.error('Failed to send signal %s to bp process %s', sig, pid)
        return -1

    log.info('%s sent to bp process %s', sig, pid)
    return 0


def run(filename, stopper,
        stop_signals=STOP_SIGNALS, restart_signals=RESTART_SIGNALS):
    signal_dict = {s: stopper for s in stop_signals + restart_signals}

    with pid_context.pid_context(filename):
        with context(**signal_dict) as signals:
            running = True
            while running:
                yield
                if not signals:
                    return
                log.info('Received signal %s', ' '.join(signals))
                running = all(s in restart_signals for s in signals)
                signals.clear()


def make_command(default_signal, help=''):
    # Handle legacy commands restart, kill, shutdown

    def add_arguments(parser):
        parser.add_argument(
            'signal', nargs='?', default=default_signal,
            choices=sorted(SIGNAL_NUMBERS),
            help='Signal to send.' + help)
        pid_context.add_arguments(parser)

    def run(args):
        return send_signal(args.signal, args.pid_filename)

    return add_arguments, run



"""
Run specified project from file or URL
"""

from .. main import project_flags
from .. project import project_runner
from .. util import signal_handler


def run(args):
    for i in signal_handler.run(args.pid_filename, project_runner.stop):
        project_runner.run(args)


def add_arguments(parser):
    parser.set_defaults(run=run)
    project_flags.add_arguments(parser)

    parser.add_argument(
        'name', nargs='*',
        help='Path project files - can be a URL or file system location')



from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    from . circle import Circle
    from . matrix import Matrix
    from . cube import Cube
    from . strip import Strip
    from . geometry.matrix import make_matrix_coord_map
    from . geometry.circle import make_circle_coord_map
    from . geometry.cube import make_cube_coord_map
    from . geometry.rotation import Rotation

    LEDCircle, LEDCube, LEDMatrix, LEDStrip = Circle, Cube, Matrix, Strip



from . import deprecated
if deprecated.allowed():  # pragma: no cover
    from .. colors import *  # noqa: F403
    from .. colors import (
        arithmetic, classic, closest_colors, conversions, gamma, juce,
        legacy_palette, make, names, palettes, printer, wheel)



from .. project import data_maker, project
from .. util import data_file, deprecated, log, pid_context

"""Common command line arguments for run and demo."""

COMPONENTS = 'driver', 'layout', 'animation'
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
ENABLE_PRESETS = False
NUMBER_TYPES = ('python',) + data_maker.NUMPY_TYPES
ADD_REDUNDANT_ARGUMENTS = True


def add_arguments(parser):
    pid_context.add_arguments(parser)
    if ADD_REDUNDANT_ARGUMENTS:
        _add_redundant_arguments(parser)

    parser.add_argument(
        '-b', '--brightness', default=None,
        help='Override project brightness value')

    parser.add_argument(
        '--dump', action='store_true',
        help='Dump the full project as DATA_FILE after loading but ' +
             'before running')

    parser.add_argument(
        '--dry_run', action='store_true',
        help='Load projects but do not run them')

    parser.add_argument(
        '-f', '--fail_on_exception', action='store_true',
        help='If true, bp fail if any subanimation fails to construct')

    if deprecated.allowed():
        parser.add_argument(
            '-g', '--gif', default='', nargs='?', help=GIF_HELP)

    parser.add_argument(
        '-i', '--ignore_exceptions', action='store_true',
        help='If true, continue running the next project if one project fails')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Use Json when dumping description data')

    parser.add_argument(
        '-m', '--movie', default='', nargs='?', help=MOVIE_HELP)

    parser.add_argument(
        '--pause', default=0, help='Time to pause between running animations')

    parser.add_argument(
        '--project_lengths', '--pl', default=None, help=PROJECT_LENGTHS_HELP)

    parser.add_argument(
        '-s', action='store_true', help='Run SimPixel at the default URL')

    parser.add_argument(
        '--simpixel', help='Run SimPixel at a specific URL')

    parser.add_argument(
        '--animation_lengths', '--at', default=None,
        help='Set run length for each animation')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')


def _add_redundant_arguments(parser):
    """
    These arguments are redundant with just using a project, and we should
    encouraging that as you don't have to learn any dumb flags!

    For example, instead of

       bp foo.yml --animation=wombat --numbers=float

    use

       bp foo.yml + '{animation: wombat, numbers: float}'

    """
    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')

    if deprecated.allowed():  # pragma: no cover
        parser.add_argument(
            '--dimensions', '--dim', default=None,
            help='DEPRECATED: x, (x, y) or (x, y, z) dimensions for project')

    parser.add_argument(
        '--shape', default=None,
        help='x, (x, y) or (x, y, z) dimensions for project')

    parser.add_argument(
        '-l', '--layout', default=None,
        help='Default layout class if no layout is specified')

    parser.add_argument(
        '--numbers', '-n', default='python', choices=NUMBER_TYPES,
        help=NUMBERS_HELP)

    parser.add_argument('-p', '--path', default=None, help=PATH_HELP)


def _make_project_flags(args):
    def get_value(name):
        value = getattr(args, name, None)
        if not value:
            return {}

        if '{' in value:
            return data_file.loads(value)

        return {'typename': value}

    project_flags = {name: get_value(name) for name in COMPONENTS}
    if args.ledtype:
        project_flags['driver']['ledtype'] = args.ledtype

    if args.brightness:
        project_flags['layout']['brightness'] = int(args.brightness)

    if args.v4:
        log.printer('*** Using v4 forward compatibility mode.')
        project_flags['numbers'] = 'float'
    elif args.numbers != 'python':
        project_flags['numbers'] = args.numbers

    if args.project_lengths is not None:
        run = project_flags.setdefault('run', {})
        run['seconds'] = float(args.project_lengths)

    if args.animation_lengths is not None:
        animation = project_flags.setdefault('animation', {})
        length = [float(i) for i in args.animation_lengths.split(',')]
        animation['length'] = length

    if deprecated.allowed():
        if args.dimensions is not None:
            deprecated.deprecated('Use --shape instead of --dimensions')
        shape = args.shape or args.dimensions
    else:
        shape = args.shape

    if shape is not None:
        shape = shape.split(',')
        try:
            project_flags['shape'] = [int(i) for i in shape]
        except:
            raise ValueError('--shape must be one to three numbers '
                             'separated by commas.')

    return project_flags


def make_project(args, *descs, root_file=None):
    project_flags = _make_project_flags(args)
    descs += (project_flags,)
    return project.project(*descs, root_file=root_file)


# Help messages.

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use loady-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

NUMBERS_HELP = """
The type of numbers that are used in color list calculations.

  `python` means to use the classic Python lists of (r, g, b) tuples.

  Anything else is a numpy type, which means that bp uses numpy arrays, which
  use much faster arithmetic.

  numpy types include:
    """ + ' '.join(data_maker.NUMPY_TYPES)

PRESET_HELP = """Filenames for preset library"""

PROJECT_LENGTHS_HELP = """\
How long to run the animation (overrides runner.seconds)"""

MOVIE_HELP = """
Write a movie file (animated GIF or mp4).

If --gif has no argument, the name of the movie file is the same as the name of
the project, with a .gif added to the end.

If it has a single string argument, then it's the name of the GIF file.

Otherwise, the argument is read as JSON and used to construct a GifWriter class.
"""

# TODO: this should go somewhere else
"""
Set defaults for projects using YAML or JSON data files.

A Project is a data file or equivalently a Python dictionary that describes
a BibliopPixel installation.  The top-level keys in a Project are called the
sections and a Project might have the following sections:

    {sections}

Often some sections of your project correspond to hardware and thus rarely
change, so the `bp default` command allows you to set defaults so you
don't have to mention these from your project at all.
"""

GIF_HELP = """
--gif/-g is a deprecated name for the --movie/-m flag."
""" + MOVIE_HELP



import itertools, weakref
from .. colors import make
from .. util import color_list, log


class Pixels:
    """
    Wrap pixels attached to a layout for easy access and
    better error reporting.
    """
    def __init__(self, builder):
        self.builder = weakref.ref(builder)

    def __getitem__(self, index):
        """
        Returns the r, g, b pixel at a location in the layout.  May only be
        called if self.is_running is true.
        """
        index = self._check_index(index)
        return self.layout.get(*index)

    def __setitem__(self, index, color):
        """
        Sets the r, g, b pixel at a location in the layout.  May only be called
        if self.is_running is true.
        """
        index = self._check_index(index)
        try:
            color = make.color(color)
        except:
            log.error('Do not understand color %s', color)
            raise
        index.append(color)
        return self.layout.set(*index)

    def clear(self):
        cl = self.layout.color_list
        color_list.Math(cl).clear(cl)

    @property
    def layout(self):
        b = self.builder()

        if not (b and b.project):
            raise ValueError('Cannot get layout before Builder has started')
        return b.project.layout

    @property
    def shape(self):
        return self.layout.shape

    def _check_index(self, index):
        if isinstance(index, int):
            index = [index]
        else:
            index = list(index)

        l1, l2 = len(self.shape), len(index)
        if l1 != l2:
            msg = 'Expected %d coordinates but got %d: %s'
            raise ValueError(msg % (l1, l2, index))

        for i, s in enumerate(self.shape):
            if index[i] < 0:
                index[i] += s
            if not (0 <= index[i] < s):
                raise IndexError('Index %d was out of range' % i)

        return index



"""
Send a restart signal to a BiblioPixel process running on this
machine.

DEPRECATED: use

.. code-block:: bash

    $ kill -hup `bpa-pid`


"""

DESCRIPTION = """
Example: ``$ bp restart``

"""

from .. util.signal_handler import make_command
add_arguments, run = make_command('SIGHUP', ' Default SIGHUP restarts bp.')
